package bufferExample

import scala.actors._
import scala.actors.Actor._
import java.util.Date

object BufferExample {
    
  abstract class BufferMessage
  case class Put(value: String) extends BufferMessage
  case object Get extends BufferMessage
  
  abstract class BufferResponse
  case class Value(nextVal: String) extends BufferResponse
  case object Stopped extends BufferResponse

  class BoundedBuffer extends Actor {
    val MaxSize = 10;
    val elt: Array[String] = new Array[String](MaxSize);
    var in = 0;          // next empty slot for new element
    var out : Int = 0;   // next element to be removed from buffer -- -1 if empty
    var count: Int = 0;  // count of elements in buffer
    
    def act() {
      loop{
        reactWithin(2000){
          case Put(value) if(count < MaxSize) =>  // add something to buffer if room
            elt(in) = value;
            in = (in + 1) % MaxSize;
            count = count+1
            Console.println("buffer has "+count+" items after Put")
          case Get if (count > 0) =>              // remove something if something there
            val value:String = elt(out);
            out = (out + 1) % MaxSize;
            count = count - 1
            reply(Value(value))
            Console.println("buffer has "+count+" items after Get")
          case TIMEOUT =>                         // Shut down the buffer if too much time
            Console.println("Buffer done");
            loop {
              reactWithin(1000){  // stay alive inform clients shut until 1 sec w/out requests
                case Get => 
                  reply(Stopped)  // sends Stopped to sender of Get
                case TIMEOUT =>
                  exit()
            }
          }
        } 
      }
    }
  }
  
  val random = new java.util.Random()
  
  // Producer tries to insert numMessages strings into buffer
  class Producer(name: String, numMessages: Int, buffer: BoundedBuffer) extends Actor {
    def log(msg: String) = Console.println(name + ": " + msg)
    val startTime = new Date().getTime()
    def act() {
      for (count <- 1 until numMessages) {
        buffer! Put(name+" item "+count)
        log("Put "+name+" item "+count +" at "+ (new Date().getTime() - startTime))
        Thread.sleep(1+random.nextInt(1000))
      }
    }
  }
  
  // Consumer tries to retrieve numMessages strings from buffer
  class Consumer(name: String, numMessages: Int, buffer: BoundedBuffer) extends Actor {
    def log(msg: String) = Console.println(name + ": " + msg)
    val startTime = new Date().getTime()
    var stopped:Boolean = false
    def act() {
      var count = 1;
      while (count <= numMessages && !stopped) {
        log("Requesting item at " + (new Date().getTime() - startTime))
        buffer!? Get match {  // !? sends message and waits for reply
          case Value(value) => 
            log(name+" received " + value +" at "+ (new Date().getTime() - startTime))
          case Stopped =>
            log("Buffer shut down so "+name+" halting")
            stopped = true
        }	
        Thread.sleep(1+random.nextInt(1000))
      }
      log("shutting down")
    }
 
  }
     
  def main(args: Array[String]) {
    val buffer = new BoundedBuffer;
    val producer1 = new Producer("Alice",25,buffer)
    val producer2 = new Producer("Bob",25,buffer)
    val consumer1 = new Consumer("Carol",20, buffer)
    val consumer2 = new Consumer("Dick",20, buffer)
    val consumer3 = new Consumer("Ellen",20, buffer)
    buffer.start
    producer1.start
    producer2.start
    consumer1.start
    consumer2.start
    consumer3.start
  } 
  

}
